/* PointerHack.c*/

/*
Changes done:
-------------
2001/12/28 - First version
2001/01/02 - Added option 'p' = Move cursor to Insertion Point (i.e. Cursor location)
2001/01/05 - Added configurable launcher key
2001/01/08 - Fixed Menu Control - now don't need to tap '1' twice :)
*/


#include "PalmOS.h"
#include "PointerHack.h"

#pragma segment 1000

Boolean MySysHandleEventTrap (EventPtr event);
void EnablePointer (AppPrefPtr Prefs);
void ShowPointer(AppPrefPtr Prefs);
void HidePointer(AppPrefPtr Prefs);
void KillPointer(AppPrefPtr Prefs);
void DrawPointer(AppPrefPtr Prefs, Int32 X, Int32 Y);
void MovePointer(AppPrefPtr Prefs, Int32 DeltaX, Int32 DeltaY);
void ProcessMenu(AppPrefPtr Prefs, Int16 MenuSelected);
void DisplayCoords(Int32 X, Int32 Y, Int32 Pos);
char *GetEventName(EventPtr event);
void initPrefs(AppPrefPtr prefP);
void savePrefs(AppPrefPtr prefP);
void loadPrefs(AppPrefPtr prefP);

Boolean MySysHandleEventTrap (EventPtr event)
{
   Boolean (*oldtrap)(EventPtr); //procedure pointer to old trap;
   UInt32 temp; //for the feature manager call type checking
   Boolean handled;
   AppPrefType MyPrefs;
   PointType ptPenPosition;   
   Coord extentX, extentY;              
   RectangleType rP;
   MenuBarPtr mbtMenu;
   EventType MyEvent;
   Int16 tmpX, tmpY;

   UInt32 ftrEnabled;
   UInt32 ftrEnableChar;
   
   FtrGet(CREATOR_ID,RESOURCE_ID,&temp); //get old trap address from HackMaster
   oldtrap=(Boolean (*)(EventPtr))temp; //set procedure pointer

   handled=false;

   //See if enabled
   if (FtrGet(CREATOR_ID, FTR_NUMBER_ENABLED, &ftrEnabled) != 0)
   {
        FtrSet(CREATOR_ID, FTR_NUMBER_ENABLED, FTR_DISABLED);
        MyPrefs.IsEnabled = false;
   }
   else
   {
        if (ftrEnabled == FTR_ENABLED)
        {
            MyPrefs.IsEnabled = true;
        }
        else
        {
            MyPrefs.IsEnabled = false;
        }
   }
   
   //Get Enable Character
   if (FtrGet(CREATOR_ID, FTR_NUMBER_ENABLECHAR, &ftrEnableChar) != 0)
   {
        loadPrefs(&MyPrefs);
        FtrSet(CREATOR_ID, FTR_NUMBER_ENABLECHAR, MyPrefs.EnableChar);
   }
   else
   {
        MyPrefs.EnableChar = ftrEnableChar;
   }
    
   if (MyPrefs.IsEnabled)
   {
       loadPrefs(&MyPrefs);
       switch (event->eType)
       {
        case keyDownEvent :
            handled = true;
            switch (event->data.keyDown.chr)
            {
                    //Normal speed
                    case 'i' : //UP
                        MovePointer(&MyPrefs, 0,-POINTER_SPEED_FAST);
                        break;

                    case chrUpArrow : //UP
                        MovePointer(&MyPrefs, 0,-POINTER_SPEED_FAST);
                        break;

                    case 'm' : //DOWN
                        MovePointer(&MyPrefs,0,POINTER_SPEED_FAST);
                        break;

                    case chrDownArrow : //DOWN
                        MovePointer(&MyPrefs,0,POINTER_SPEED_FAST);
                        break;

                    case 'j' : //LEFT
                        MovePointer(&MyPrefs,-POINTER_SPEED_FAST,0);
                        break;

                    case chrLeftArrow : //LEFT
                        MovePointer(&MyPrefs,-POINTER_SPEED_FAST,0);
                        break;

                    case 'k' : //RIGHT
                        MovePointer(&MyPrefs,POINTER_SPEED_FAST,0);
                        break;

                    case chrRightArrow : //RIGHT
                        MovePointer(&MyPrefs,POINTER_SPEED_FAST,0);
                        break;

                    case 'o' : //UP & RIGHT
                        MovePointer(&MyPrefs,POINTER_SPEED_FAST,-POINTER_SPEED_FAST);
                        break;

                    case ',' : //DOWN & RIGHT
                        MovePointer(&MyPrefs,POINTER_SPEED_FAST,POINTER_SPEED_FAST);
                        break;

                    case 'u' : //UP & LEFFT
                        MovePointer(&MyPrefs,-POINTER_SPEED_FAST,-POINTER_SPEED_FAST);
                        break;                       

                    case 'n' : //DOWN & LEFT
                        MovePointer(&MyPrefs,-POINTER_SPEED_FAST,POINTER_SPEED_FAST);
                        break;

                    //Fine Tune
                    case 'I' : //UP
                        MovePointer(&MyPrefs, 0,-POINTER_SPEED_SLOW);
                        break;

                    case 'M' : //DOWN
                        MovePointer(&MyPrefs,0,POINTER_SPEED_SLOW);
                        break;

                    case 'J' : //LEFT
                        MovePointer(&MyPrefs,-POINTER_SPEED_SLOW,0);
                        break;

                    case 'K' : //RIGHT
                        MovePointer(&MyPrefs,POINTER_SPEED_SLOW,0);
                        break;

                    case 'O' : //UP & RIGHT
                        MovePointer(&MyPrefs,POINTER_SPEED_SLOW,-POINTER_SPEED_SLOW);
                        break;

                    case '<' : //DOWN & RIGHT
                        MovePointer(&MyPrefs,POINTER_SPEED_SLOW,POINTER_SPEED_SLOW);
                        break;

                    case 'U' : //UP & LEFT
                        MovePointer(&MyPrefs,-POINTER_SPEED_SLOW,-POINTER_SPEED_SLOW);
                        break;

                    case 'N' : //DOWN & LEFT
                        MovePointer(&MyPrefs,-POINTER_SPEED_SLOW,POINTER_SPEED_SLOW);
                        break;

                    case 'p' : //Go to Insertion Point (i.e. Cursor location)
                        InsPtGetLocation(&tmpX, &tmpY);
                        WinGetWindowBounds(&rP);
                        DrawPointer(&MyPrefs, tmpX - rP.topLeft.x, tmpY - rP.topLeft.y);                        
                        break;


                    //Miscellaneous
                    case 't' : //Top
                        DrawPointer(&MyPrefs,MyPrefs.CurrentX, 7);
                        break;

                    case 'b' : //Bottom
                        DrawPointer(&MyPrefs,MyPrefs.CurrentX, MyPrefs.ScreenMaxY - 10);
                        break;

                    case 'l' : //Left
                        DrawPointer(&MyPrefs,10, MyPrefs.CurrentY);
                        break;

                    case 'r' : //Right
                        DrawPointer(&MyPrefs,MyPrefs.ScreenMaxX - 7, MyPrefs.CurrentY);
                        break;

                    case 'c' : //Centre
                        DrawPointer(&MyPrefs,MyPrefs.ScreenMaxX/2 , MyPrefs.ScreenMaxY/2);
                        break;

                    case 'z' : //Toggle Menu
                        HidePointer(&MyPrefs);
                        KillPointer(&MyPrefs);
                        MyEvent.eType = penDownEvent;
                        MyEvent.penDown = true;
                        MyEvent.screenX = 12;
                        MyEvent.screenY = 205;
                        EvtAddEventToQueue(&MyEvent);
                        MyEvent.eType = penUpEvent;
                        MyEvent.penDown = false;
                        MyEvent.screenX = 12;
                        MyEvent.screenY = 205;
                        MyEvent.data.penUp.start.x = 12;
                        MyEvent.data.penUp.start.y = 205;
                        MyEvent.data.penUp.end.x = 12;
                        MyEvent.data.penUp.end.y = 205;
                        EvtAddEventToQueue(&MyEvent);
                        break;
                        
                    case '1' : //Pull Down Menu 1
                        HidePointer(&MyPrefs);
                        KillPointer(&MyPrefs);
                        ProcessMenu(&MyPrefs, 1);
                        break;
                        
                    case '2' : //Pull Down Menu 2
                        HidePointer(&MyPrefs);
                        KillPointer(&MyPrefs);
                        ProcessMenu(&MyPrefs, 2);
                        break;
                        
                    case '3' : //Pull Down Menu 3
                        HidePointer(&MyPrefs);
                        KillPointer(&MyPrefs);
                        ProcessMenu(&MyPrefs, 3);
                        break;
                        
                    case '4' : //Pull Down Menu 4
                        HidePointer(&MyPrefs);
                        KillPointer(&MyPrefs);
                        ProcessMenu(&MyPrefs, 4);
                        break;
                    
                    case '5' : //Pull Down Menu 5
                        HidePointer(&MyPrefs);
                        KillPointer(&MyPrefs);
                        ProcessMenu(&MyPrefs, 5);
                        break;
                    
                    case '6' : //Pull Down Menu 6
                        HidePointer(&MyPrefs);
                        KillPointer(&MyPrefs);
                        ProcessMenu(&MyPrefs, 6);
                        break;
                    
                    case '7' : //Pull Down Menu 7
                        HidePointer(&MyPrefs);
                        KillPointer(&MyPrefs);
                        ProcessMenu(&MyPrefs, 7);
                        break;
                    
                    case '8' : //Pull Down Menu 8
                        HidePointer(&MyPrefs);
                        KillPointer(&MyPrefs);
                        ProcessMenu(&MyPrefs, 8);
                        break;
                    
                    case '9' : //Pull Down Menu 9
                        HidePointer(&MyPrefs);
                        KillPointer(&MyPrefs);
                        ProcessMenu(&MyPrefs, 9);
                        break;
                    
                    case '0' : //Pull Down Menu 10
                        HidePointer(&MyPrefs);
                        KillPointer(&MyPrefs);
                        ProcessMenu(&MyPrefs, 10);
                        break;
                    
                    case 's' : //SELECT single click
                        HidePointer(&MyPrefs);
                        KillPointer(&MyPrefs);
                        //Pen Down
                        EvtFlushPenQueue();
                        WinGetWindowBounds(&rP);
                        //Set current position for evtGetPen
                        ptPenPosition.x = MyPrefs.DigMaxX -  
                           (80-(((((80-(rP.topLeft.x + MyPrefs.CurrentX))*100)/80)*MyPrefs.MagDigX)/100));
                        ptPenPosition.y = MyPrefs.DigMaxY - 
                           (80-(((((80-(rP.topLeft.y + MyPrefs.CurrentY))*100)/80)*MyPrefs.MagDigY)/100));
                        EvtEnqueuePenPoint(&ptPenPosition);   
                        //Pen Up
                        ptPenPosition.x = -1;
                        ptPenPosition.y = -1;
                        EvtEnqueuePenPoint(&ptPenPosition);   
                        break;

                    case '^' : //SELECT Outside Window 
                        HidePointer(&MyPrefs);
                        KillPointer(&MyPrefs);
                        //Pen Down
                        EvtFlushPenQueue();
                        WinGetWindowBounds(&rP);
                        ptPenPosition.x = MyPrefs.DigMaxX - rP.topLeft.x; 
                        ptPenPosition.y = MyPrefs.DigMaxY - (rP.topLeft.y - 2);
                        EvtEnqueuePenPoint(&ptPenPosition);
                        //Pen Up
                        ptPenPosition.x = -1;
                        ptPenPosition.y = -1;
                        EvtEnqueuePenPoint(&ptPenPosition);   
                        break;

                    case '.' : //DISABLE
                        HidePointer(&MyPrefs);
                        KillPointer(&MyPrefs);
                        //KillMenu(&MyPrefs);
                        break;


                    default : 
                        break;
                } //switch (event->data.keyDown.chr)
                
                if (event->data.keyDown.chr == MyPrefs.EnableChar) //DO ACTUAL CHARACTER - NOT ENABLE
                {
                        HidePointer(&MyPrefs);
                        KillPointer(&MyPrefs);
                        handled = false;
                }
                break;
       
        case appStopEvent :
            HidePointer(&MyPrefs);
            KillPointer(&MyPrefs);
            //KillMenu(&MyPrefs);
            handled = false;
            break;

        case tblSelectEvent :
            HidePointer(&MyPrefs);
            KillPointer(&MyPrefs);
            break;

        case ctlSelectEvent :
            HidePointer(&MyPrefs);
            KillPointer(&MyPrefs);
            break;

        case lstSelectEvent :
            HidePointer(&MyPrefs);
            KillPointer(&MyPrefs);
            break;

        case popSelectEvent :
            HidePointer(&MyPrefs);
            KillPointer(&MyPrefs);
            break;

        case daySelectEvent :
            HidePointer(&MyPrefs);
            KillPointer(&MyPrefs);
            break;

        case frmTitleSelectEvent :
            HidePointer(&MyPrefs);
            KillPointer(&MyPrefs);
            break;

        case ctlRepeatEvent :
            HidePointer(&MyPrefs);
            KillPointer(&MyPrefs);
            break;

        case sclRepeatEvent :
            HidePointer(&MyPrefs);
            KillPointer(&MyPrefs);
            break;

        case winExitEvent :
            HidePointer(&MyPrefs);
            KillPointer(&MyPrefs);
            break;

        default:
            break;

       } //switch (event->eType)

    savePrefs(&MyPrefs);
   } //if (MyPrefs.IsEnabled)

   else //Not enabled
   {
        if (event->data.keyDown.chr == MyPrefs.EnableChar && event->eType == keyDownEvent)
        {
            loadPrefs(&MyPrefs);
            initPrefs(&MyPrefs);

            WinGetWindowExtent(&extentX, &extentY);
            MyPrefs.ScreenMaxX = extentX;
            MyPrefs.ScreenMaxY = extentY;

            EnablePointer(&MyPrefs);
            ShowPointer(&MyPrefs);

            handled = true;
            savePrefs(&MyPrefs);
        }
   } //else: if (MyPrefs.IsEnabled)

   if (MyPrefs.IsEnabled)
   {
        FtrSet(CREATOR_ID, FTR_NUMBER_ENABLED, FTR_ENABLED);
   }
   else
   {
        FtrSet(CREATOR_ID, FTR_NUMBER_ENABLED, FTR_DISABLED);
   }
       
   if (!handled) //if you didn't handle it, call the old routine
   { 
       handled=oldtrap(event);
   } //if (!handled)
   return handled;
}
                            
void EnablePointer (AppPrefPtr Prefs)
{
    Prefs->IsEnabled = true;
}

void ShowPointer (AppPrefPtr Prefs)
{
    if (Prefs->IsEnabled)
    {
        DrawPointer(Prefs, Prefs->CurrentX, Prefs->CurrentY);
        Prefs->IsVisible = true;
    }
}

void HidePointer (AppPrefPtr Prefs)
{
    if (Prefs->IsEnabled)
    {
        Prefs->IsVisible = false;
        DrawPointer(Prefs, Prefs->CurrentX, Prefs->CurrentY); //Turn off
    }
}

void KillPointer (AppPrefPtr Prefs)
{
    Prefs->IsEnabled = false;
}

void DrawPointer(AppPrefPtr Prefs, Int32 X, Int32 Y)
{
    if (Prefs->IsVisible)  //Remove old first
    {
        WinInvertLine (Prefs->CurrentX + 1, Prefs->CurrentY, Prefs->CurrentX + POINTER_SIZE - 1, Prefs->CurrentY + 8); 
        WinInvertLine (Prefs->CurrentX, Prefs->CurrentY, Prefs->CurrentX + 8, Prefs->CurrentY + POINTER_SIZE - 1); 
        WinInvertLine (Prefs->CurrentX + POINTER_SIZE, Prefs->CurrentY + 9, Prefs->CurrentX + 9, Prefs->CurrentY + POINTER_SIZE); 
    }

    if (X > Prefs->ScreenMaxX) X = Prefs->ScreenMaxX;
    if (X < 1) X = 1;
    if (Y > Prefs->ScreenMaxY) Y = Prefs->ScreenMaxY;
    if (Y < 1) Y = 1;

    Prefs->CurrentX = X;
    Prefs->CurrentY = Y;
    
    WinInvertLine (Prefs->CurrentX + 1, Prefs->CurrentY, Prefs->CurrentX + POINTER_SIZE - 1, Prefs->CurrentY + 8); 
    WinInvertLine (Prefs->CurrentX, Prefs->CurrentY, Prefs->CurrentX + 8, Prefs->CurrentY + POINTER_SIZE - 1); 
    WinInvertLine (Prefs->CurrentX + POINTER_SIZE, Prefs->CurrentY + 9, Prefs->CurrentX + 9, Prefs->CurrentY + POINTER_SIZE); 
}

void MovePointer(AppPrefPtr Prefs, Int32 DeltaX, Int32 DeltaY)
{
    DrawPointer(Prefs, Prefs->CurrentX + DeltaX, Prefs->CurrentY + DeltaY);
}

void ProcessMenu(AppPrefPtr Prefs, Int16 MenuSelected)
{
    MenuBarPtr mbtMenu;
    EventType MyEvent;
    FormPtr ftMenu;
    UInt16 mnuError;
    
    mbtMenu = MenuGetActiveMenu();

    if (mbtMenu == NULL) //If not active, activate it.
    {
        ftMenu = FrmGetActiveForm();
        if (ftMenu->menuRscId != 0)
        {
            mbtMenu = MenuInit(ftMenu->menuRscId);
            MenuSetActiveMenu(mbtMenu);
            mbtMenu = MenuGetActiveMenu();
            if (mbtMenu && MenuSelected > 0 && MenuSelected <= mbtMenu->numMenus)
            { 
                mbtMenu->curMenu = MenuSelected - 1; //zero-based
                mbtMenu->attr.visible = false; //zero-based
                mbtMenu->curItem = noMenuItemSelection;
                
                MyEvent.eType = keyDownEvent;
                MyEvent.data.keyDown.chr = vchrMenu;
                MyEvent.data.keyDown.modifiers = 0;
                EvtAddEventToQueue(&MyEvent);
                
                MyEvent.eType = keyDownEvent;
                MyEvent.data.keyDown.chr = Prefs->EnableChar;
                MyEvent.data.keyDown.modifiers = 0;
                EvtAddEventToQueue(&MyEvent);
            }
        }
    }
    else
    if (mbtMenu && !mbtMenu->attr.visible && !mbtMenu->attr.commandPending)
    {
          if (MenuSelected > 0 && MenuSelected <= mbtMenu->numMenus)
          {
                mbtMenu->curMenu = MenuSelected - 1; //zero-based
                mbtMenu->curItem = noMenuItemSelection;
                
                MyEvent.eType = keyDownEvent;
                MyEvent.data.keyDown.chr = vchrMenu;
                MyEvent.data.keyDown.modifiers = 0;
                EvtAddEventToQueue(&MyEvent);
                
                MyEvent.eType = keyDownEvent;
                MyEvent.data.keyDown.chr = Prefs->EnableChar;
                MyEvent.data.keyDown.modifiers = 0;
                EvtAddEventToQueue(&MyEvent);
          }
    }   
    
}


void DisplayCoords(Int32 X, Int32 Y, Int32 Pos)
{
    RectangleType rP;
    char * scrX = " ";
    char * scrY = " ";
    char * newString = " ";
    Int32 yPos = 0;
        
    StrIToA(scrX, X);    
    StrIToA(scrY, Y);    
    StrCopy(newString, scrX);
    StrCat(newString, ",");
    StrCat(newString, scrY);

    yPos = 80 + (Pos * 20);
    rP.topLeft.x = 19;
    rP.topLeft.y = yPos - 1;
    rP.extent.x = 40;
    rP.extent.y = 15;
    WinEraseRectangle(&rP,0);
    WinInvertChars (newString, StrLen(newString),20, yPos);
}

char *GetEventName(EventPtr event)
{
    char *EventName;
    
    switch (event->eType)
    {
        
        case nilEvent :
            EventName = "nilEvent";
            break;
        case penDownEvent :
            EventName = "penDownEvent";
            break;
        case penUpEvent :
            EventName = "penUpEvent";
            break;
        case penMoveEvent :
            EventName = "penMoveEvent";
            break;
        case keyDownEvent :
            EventName = "keyDownEvent";
            break;
        case winEnterEvent :
            EventName = "winEnterEvent";
            break;
        case winExitEvent :
            EventName = "winExitEvent";
            break;
        case ctlEnterEvent :
            EventName = "ctlEnterEvent";
            break;
        case ctlExitEvent :
            EventName = "ctlExitEvent";
            break;
        case ctlSelectEvent :
            EventName = "ctlSelectEvent";
            break;
        case ctlRepeatEvent :
            EventName = "ctlRepeatEvent";
            break;
        case lstEnterEvent :
            EventName = "lstEnterEvent";
            break;
        case lstSelectEvent :
            EventName = "lstSelectEvent";
            break;
        case lstExitEvent  :
            EventName = "lstExitEvent ";
            break;
        case popSelectEvent :
            EventName = "popSelectEvent";
            break;
        case fldEnterEvent :
            EventName = "fldEnterEvent";
            break;
        case fldHeightChangedEvent :
            EventName = "fldHeightChangedEvent";
            break;
        case fldChangedEvent :
            EventName = "fldChangedEvent";
            break;
        case tblEnterEvent :
            EventName = "tblEnterEvent";
            break;
        case tblSelectEvent :
            EventName = "tblSelectEvent";
            break;
        case daySelectEvent :
            EventName = "daySelectEvent";
            break;
        case menuEvent :
            EventName = "menuEvent";
            break;
        case appStopEvent :
            EventName = "appStopEvent";
            break;
        case frmLoadEvent :
            EventName = "frmLoadEvent";
            break;
        case frmOpenEvent :
            EventName = "frmOpenEvent";
            break;
        case frmGotoEvent :
            EventName = "frmGotoEvent";
            break;
        case frmUpdateEvent :
            EventName = "frmUpdateEvent";
            break;
        case frmSaveEvent :
            EventName = "frmSaveEvent";
            break;
        case frmCloseEvent :
            EventName = "frmCloseEvent";
            break;
        case frmTitleEnterEvent :
            EventName = "frmTitleEnterEvent";
            break;
        case frmTitleSelectEvent :
            EventName = "frmTitleSelectEvent";
            break;
        case tblExitEvent :
            EventName = "tblExitEvent";
            break;
        case sclEnterEvent :
            EventName = "sclEnterEvent";
            break;
        case sclExitEvent :
            EventName = "sclExitEvent";
            break;
        case sclRepeatEvent :
            EventName = "sclRepeatEvent";
            break;
        case tsmFepModeEvent :
            EventName = "tsmFepModeEvent";
            break;
        case menuCmdBarOpenEvent :
            EventName = "menuCmdBarOpenEvent";
            break;
        case menuOpenEvent :
            EventName = "menuOpenEvent";
            break;
        case menuCloseEvent :
            EventName = "menuCloseEvent";
            break;
        case frmGadgetEnterEvent :
            EventName = "frmGadgetEnterEvent";
            break;
        case frmGadgetMiscEvent :
            EventName = "frmGadgetMiscEvent";
            break;
        case firstINetLibEvent :
            EventName = "firstINetLibEvent";
            break;
        case firstWebLibEvent :
            EventName = "firstWebLibEvent";
            break;
        case firstUserEvent :
            EventName = "firstUserEvent";
            break;
        case tsmConfirmEvent :
            EventName = "tsmConfirmEvent";
            break;
        case tsmFepButtonEvent :
            EventName = "tsmFepButtonEvent";
            break;
        default : 
            EventName = "Unknown Event !";
            break;
    }   
    
    return EventName;
    
}

void initPrefs(AppPrefPtr prefP)
{
    prefP->IsEnabled = false;
    prefP->IsVisible = false;
    prefP->ScreenMaxX = 80;
    prefP->ScreenMaxY = 80;
    prefP->CurrentX = 10;
    prefP->CurrentY = 7;

    savePrefs(prefP);
}

void savePrefs(AppPrefPtr prefP)
{
    PrefSetAppPreferences(CREATOR_ID, PREF_ID, PREF_VERSION, prefP, PREF_SIZE, PREF_SAVED);
}


void loadPrefs(AppPrefPtr prefP)
{
    Int32 found;
    UInt16 prefsSize = PREF_SIZE;

    found = PrefGetAppPreferences(CREATOR_ID, PREF_ID, prefP, &prefsSize, PREF_SAVED);

    if ((found == noPreferenceFound) || (prefsSize != PREF_SIZE))
    {
        initPrefs(prefP);
    }
}
